`timescale 1ns / 1ps

//////////////////////////////////////////////////////////////////////////////////
//
//	Code written by: Jeff Casarona
//  Date: 10/15/2014
//  Property of UMD ECE Tech Ops
//
//		This code will initialize the LCD Screen for the Nexys 2 FPGA
//		and then sit in a endless loop, where it will give control over
//		to the "write_driver.v" code. The "JA" inout port acts as a bi-
//		directional bus. This means that in any case where you are not
//		actively driving the JA pins with data, you should hold them to
//		high impedance. 
//
//////////////////////////////////////////////////////////////////////////////////

module LCD_DRIVER(input clk, input reset, input [255:0] lcd_input, output RS, output R_W, output e, inout [7:0] JA
    );

	parameter wait_20ms = 22'h3F4240;
	parameter wait_39us = 22'h79E;
	parameter wait_1_53ms = 22'h12AD4;

	// Parameters for the statemachine
	localparam START=3'b000, SET_FUNC=3'b001, 
					SET_DISP=3'b010, CLEAR_DISP=3'b011, MODE_SET=3'b111, SET_HOME=3'b110,
					END_INIT=3'b100;
	
	wire RS_int;
	wire R_W_int;
	wire e_out;
	wire done_flag;
	wire [255:0] lcd_data;
	assign lcd_data = lcd_input;
	//Write Driver: This enables the writing of address locations
	// and of character data to the LCD
	write_driver wr_drv (
    .clk(clk), 
    //.switches(switches), 
    .reset(reset), 
	 .flag(done_flag),
	 .lcd_data(lcd_data),
    .e(e_out), 
    .RS(RS_int), 
    .RW(R_W_int), 
    .data(JA)
    );

	reg [2:0] state, nxt_state;
	reg [21:0] count;
	reg clp_en_clk;
	
	//Standard FSM Setup Code
	always @ (posedge clk) begin
		if(reset)
			state<=START;
		else
			state<=nxt_state;
	end 
	
	//Do next state and output updates
	// these are asynchronous
	always @ (count, state, reset) begin
		if(reset) begin
			nxt_state = START;
		end
		else begin
		case(state)
			START:
				// Wait 80ms
				nxt_state = (count == wait_20ms) ? SET_FUNC : START;
			SET_FUNC:
				// Wait (some amount of)us
				nxt_state=(count == wait_39us) ? SET_DISP : SET_FUNC;
			SET_DISP: 
				// Wait (some amount of)us
				nxt_state=(count == wait_39us) ? CLEAR_DISP : SET_DISP;
			CLEAR_DISP:
				// Wait (some amount of)ms
				nxt_state=(count == wait_1_53ms) ? MODE_SET : CLEAR_DISP;
			MODE_SET:
				// Wait (some amount of)us
				nxt_state= (count == wait_39us) ? SET_HOME : MODE_SET;
			SET_HOME:
				// Wait (some amount of)ms
				nxt_state = (count == wait_1_53ms) ? END_INIT : SET_HOME;
			END_INIT:
				nxt_state = END_INIT;
			default: begin 
				nxt_state=START;
				end
		endcase
		end
	end
	

	
	
	// Do Counting and Enable line
	// Note: For enable line, wait 40 ns before setting high,
	//		and hold this value for 220 ns before setting low
	always @ (posedge clk) begin
	if (reset) begin
		clp_en_clk <= 0;
		count <= 0;
		clp_en_clk <= 0;
	end
	else begin
		case(state)
			START: begin
				count <= (count == wait_20ms) ? 22'h0 : count + 1; // 20ms
				clp_en_clk <= 0;
				end
			SET_FUNC: begin
				count <= (count == wait_39us) ? 22'h0 : count + 1; // 37us
				clp_en_clk <= (count > 22'h1E && count < 22'h3C ) ? 1 : 0;
				end
			SET_DISP: begin
				count <= (count == wait_39us) ? 22'h0 : count + 1; // 37us
				clp_en_clk <= (count > 21'h1E && count < 22'h3C ) ? 1 : 0;
				end
			CLEAR_DISP: begin
				count <= (count == wait_1_53ms) ? 22'h0 : count + 1; // 1.52ms
				clp_en_clk <= (count > 22'h1E && count < 22'h3C ) ? 1 : 0;
				end
			MODE_SET: begin
				count <= (count == wait_39us) ? 22'h0 : count + 1;
				clp_en_clk <= (count > 22'h1E && count < 22'h3C) ? 1 : 0;
				end
			SET_HOME: begin
				count <= (count == wait_1_53ms) ? 22'h0 : count + 1;
				clp_en_clk <= (count > 22'h1E && count < 22'h3C ) ? 1 : 0;
				end
			END_INIT: begin
				count <= count + 1;
			end
			default: begin
				count <= 22'h0;
				clp_en_clk <= 0;
				end
		endcase
	end
	end
	
	assign done_flag = (state == END_INIT);
	//Enable line needs to update with clp enable
	assign e = (state == END_INIT) ? e_out : clp_en_clk;
	//These commands are in the datasheet for their respective function
	// Then set the output of this to high impedance (Z)
	assign JA = (state == SET_FUNC) ? 8'h3C : (state == SET_DISP) ? 8'h0F :
			(state == CLEAR_DISP) ? 8'h01 : (state == MODE_SET) ? 8'h06 : 
			(state == SET_HOME) ? 8'h02 : 8'hZ;//8'h41;
	// Hold low until the end of the initialization, where we pass
	//  control over to the write_driver.v
	assign RS = (state == END_INIT) ? RS_int : 0;
	assign R_W = (state == END_INIT) ? R_W_int : 0;	
	//assign R_W = 0;
	

endmodule

//////////////////////////////////////////////////////////////////////////////////
//
//	Code written by: Jeff Casarona
//  Date: 10/15/2014
//  Property of UMD ECE Tech Ops
//
// This code is the main address and data driver to write characters to
// the LCD screen. Commands are triggered by pressing either button 1 or
// button 2, which will either set the address to write a character, or 
// write data to the LCD screen. Operation of the board:
//
//  After startup, blinking cursor will show at Address 0x00 of the board.
//  From here, "btn0" will reset the board by clearing the display and
//  setting the cursor back to the "Home" position, which is 0x00.
//
//  Pressing "btn1" will write data to the LCD screen. A standard ASCII
//  character scheme is followed for this. By default, writing a character
//  will increment your cursor to the next location. This makes it easy to
//  write a long string of characters.
//  
//  Pressing "btn2" will write the address to DDRAM. This will move your 
//  cursor to the corresponding location on the LCD Screen. The LCD
//  screen has 16 characters on two rows. The visible address locations
//  range from 0x00 to 0x0F for the top row, and 0x40 to 0x4F for the 
//  bottom row.
//
//  All data is input from the switches. For example, after programming the
//  board, entering 0x41 on the switches and pressing "btn1" will print a 
//  single "A" character in the top left corner of the display, and move
//  the cursor to the right by one. 
//
//////////////////////////////////////////////////////////////////////////////////

module write_driver(input clk, input reset, input flag, input [255:0] lcd_data,
		output e, output RS, output RW, inout [7:0] data
    );
	 
	 localparam START=3'b000, BUSY_A=3'b001, WR_ADDR=3'b010, BUSY_D=3'b011, WR_DATA=3'b100, SH_DATA=3'b101, INIT=3'b111;
	 parameter MAX_COUNT = 16'h3FF;
	 parameter SLOW_COUNT = 16'hFFFF;
	 //parameter PE_E = 10'h1E;
	 //parameter NE_E = 10'h3C;
	 parameter PE_E = 16'h1E;
	 parameter NE_E = 16'h3C;
	 parameter ZERO_COUNT = 16'h0000;
	 
	 //reg [11:0] sh_data_btn;
	 //reg [11:0] sh_addr_btn;
	 reg [2:0] state, nxt_state;
	 reg [15:0] count;
	 reg clp_en_clk;
	 reg ack_rcv;
	 reg [255:0] data_buffer;
	 //reg [15:0] slow_count;
	 reg [4:0] addr_count;
	 
	 // Code used to generate a debounced button so only one character is
	 // printed at a time.
	 wire pe_data_btn;
	 
	 
	 always @ (posedge clk) begin
		if (reset)
			state <= START;
		else
			state <= nxt_state;
	 end
	 
//////////////////////////////////////////////////////////////////////////////////
//  
//  This is the state machine to write the address or data to the LCD screen.
//  This process is a 3 step process
//
//  1. Wait for a button press (btn1 or btn2) to occur. Any enable signal should
// 		should be fine here
//  2. Check the busy flag from the LCD to make sure that the screen is not
//			doing anything currently. Do not leave until the LCD screen is !busy
//  3. Write the data or address from the switches to the LCD screen
//
//////////////////////////////////////////////////////////////////////////////////
	 always @ ( * ) begin
		nxt_state = START;
		if(reset) begin
			nxt_state = START;
		end
		else begin
			case(state)
				INIT: nxt_state = (flag) ? START : INIT;
				START: nxt_state = (pe_data_btn) ? (addr_count == 5'b00000 || addr_count == 5'b10000) ? BUSY_A : BUSY_D : START;//(pe_data_btn) ? BUSY_D : nxt_state;
				BUSY_A: nxt_state = (ack_rcv) ? WR_ADDR : BUSY_A;
				// Delay the write states to slow down the board. This is not necessary
				WR_ADDR: nxt_state = (pe_data_btn) ? BUSY_D : WR_ADDR;
				BUSY_D: nxt_state = (ack_rcv) ? WR_DATA : BUSY_D;
				WR_DATA: nxt_state = (pe_data_btn) ? SH_DATA : WR_DATA;
				SH_DATA: nxt_state = START;
				default: nxt_state = START;
			endcase
		end
	 end
	 
	 always @ (posedge clk) begin
		if(reset) begin
			count <= ZERO_COUNT;
			clp_en_clk <= 0;
			data_buffer <= lcd_data;
			addr_count <= 0;
			ack_rcv <= 0;
		end
		else begin
			case(state)
				START: begin
					count <= count + 1;
					clp_en_clk <= 0;
					data_buffer <= (addr_count == 0) ? lcd_data : data_buffer;
					ack_rcv <= 0;
				end
				BUSY_A: begin
					count <= count + 1;
					clp_en_clk <= (count > PE_E && count < NE_E ) ? 1 : 0;
					ack_rcv <= (clp_en_clk && (data[7] == 0)) ? 1 : ack_rcv;
				end
				WR_ADDR: begin
					count <= count + 1;
					clp_en_clk <= (count > PE_E && count < NE_E ) ? 1 : 0;
					ack_rcv <= 0;
				end
				BUSY_D: begin
					count <= count + 1;
					clp_en_clk <= (count > PE_E && count < NE_E ) ? 1 : 0;
					ack_rcv <= (clp_en_clk && (data[7] == 0)) ? 1 : ack_rcv;
				end
				WR_DATA: begin
					count <= count + 1;
					clp_en_clk <= (count > PE_E && count < NE_E ) ? 1 : 0;
					ack_rcv <= 0;
				end
				SH_DATA:	begin 
					//data_buffer <= { data_buffer[7:0], data_buffer[255:8] };
					data_buffer <= { data_buffer[247:0], data_buffer[255:248] };
					addr_count <= addr_count + 1;
					count <= 0;
				end
				default: begin
					count <= ZERO_COUNT;
					clp_en_clk <= 0;
					addr_count <= 0;
					ack_rcv <= 0;
				end
			endcase
		end
	 end
	 
	 
	 assign e = clp_en_clk;
	 assign RS = (state == WR_DATA) ? 1 : 0;
	 assign RW = ((state == BUSY_A) || (state == BUSY_D)) ? 1 : 0;
	 assign data = (state == WR_DATA) ? data_buffer[255:248] : (state == WR_ADDR && addr_count == 0) ? 
		{1'b1, 7'b0000000} : (state == WR_ADDR && addr_count == 16) ? {1'b1, 7'b1000000} : 8'bZ;
	 
	 // Debounce logic: If 12'b011111111111 is shown, then that means we have a 
	 //  stable 0->1 transition, and you should move to the next state
	 // This debounce logic is not perfect. In order to use the buttons
	 //  and have perfect accuracy, debounce register needs to be larger
	 assign pe_data_btn = (count == 16'hFFFF);
endmodule
